#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 25 13:39:56 2019

@author: salbreg/williamsonj
"""

import random
import copy
import numpy as np
import matplotlib.pyplot as plt
from Tissue import tissue


class run_population_growth(object):
    """ class that run a simulation for population growth"""
    DIVIDING_CELL_TYPES = {0:1, 1:0, 2:0}#cells of type 0 divide, cells of type 1 or 2 (SOPs) do not
    DELTA_T = 2.5/60.0 #as in experiment, delta_t is taken to be 2.5 minutes
    TOTAL_TIME = 40 #total time in hours for the simulation
    def __init__(self,
                 file_suffix,
                 STARTING_TIME,
                 SECOND_PHASE_STARTING_TIME,
                 INITIAL_NUMBER_CELLS,
                 INITIAL_AGE,
                 MEAN_CELL_CYCLE_BEGINNING,
                 STD_CELL_CYCLE_BEGINNING,
                 MEAN_CELL_CYCLE_TRANSIENT,
                 STD_CELL_CYCLE_TRANSIENT,
                 SWITCH_TIME_ALPHA,
                 HILL_COEFF_ALPHA,
                 SWITCH_TIME_P,
                 HILL_COEFF_P,
                 TRANSIENT_PHASE_STARTING_TIME,
                 SWITCH_TIME_SOPs_ON,
                 SWITCH_TIME_SOPs_OFF,
                 PROBABILITY_SOP,
                 COV_CELL_CYCLE,
                 data_life_times_over_time,
                 MU,
                 SIGMA,
                 RHO_CORRELATION_COEFFICIENT,
                 MODE,
                 PAUSE_ON,
                 PAUSE_OFF,
                 AGE_ON,
                 AGE_OFF,
                 folder_name_results):
        """class initialization
        MODE can be 1 (mean of cell cycle time distribution in second phase
        is taken from a fit to experimental data  and std is proportional to mean)
        or can be 2 ( cell cycle time distribution has a fixed mean and std)"""
        self.file_suffix = file_suffix
        self.STARTING_TIME = STARTING_TIME
        self.SECOND_PHASE_STARTING_TIME = SECOND_PHASE_STARTING_TIME
        self.INITIAL_NUMBER_CELLS = INITIAL_NUMBER_CELLS
        self.INITIAL_AGE = INITIAL_AGE
        self.MEAN_CELL_CYCLE_BEGINNING = MEAN_CELL_CYCLE_BEGINNING
        self.STD_CELL_CYCLE_BEGINNING = STD_CELL_CYCLE_BEGINNING
        self.MEAN_CELL_CYCLE_TRANSIENT = MEAN_CELL_CYCLE_TRANSIENT
        self.STD_CELL_CYCLE_TRANSIENT = STD_CELL_CYCLE_TRANSIENT
        self.SWITCH_TIME_ALPHA = SWITCH_TIME_ALPHA
        self.HILL_COEFF_ALPHA = HILL_COEFF_ALPHA
        self.SWITCH_TIME_P = SWITCH_TIME_P
        self.HILL_COEFF_P = HILL_COEFF_P
        self.TRANSIENT_PHASE_STARTING_TIME = TRANSIENT_PHASE_STARTING_TIME
        self.SWITCH_TIME_SOPs_ON = SWITCH_TIME_SOPs_ON
        self.SWITCH_TIME_SOPs_OFF = SWITCH_TIME_SOPs_OFF
        self.PROBABILITY_SOP = PROBABILITY_SOP
        self.COV_CELL_CYCLE = COV_CELL_CYCLE
        self.data_life_times_over_time = data_life_times_over_time
        self.MU = MU
        self.SIGMA = SIGMA
        self.RHO_CORRELATION_COEFFICIENT = RHO_CORRELATION_COEFFICIENT
        self.MODE = MODE
        self.PAUSE_ON = PAUSE_ON
        self.PAUSE_OFF = PAUSE_OFF
        self.AGE_ON = AGE_ON
        self.AGE_OFF = AGE_OFF
        self.folder_name_results = folder_name_results
        #initial cells have an age INITIAL_AGE
        self.INITIAL_AGES = np.zeros(self.INITIAL_NUMBER_CELLS) 
        self.INITIAL_AGES.fill(self.INITIAL_AGE) 
        #initial cells have cell type 0 by default
        self.INITIAL_TYPES = np.zeros(self.INITIAL_NUMBER_CELLS) 
        #generate a set of initial cell cycle times
        self.INITIAL_CELL_CYCLE_TIMES = [self.cell_cycle_generator_bivariate(0)[1]
                                         for i
                                         in range(len(self.INITIAL_TYPES))]
 
#note: beyond the boundaries of discrete values,
# the interp function returns the closest value
    def mean_cell_cycle_function(self, simulation_time):
        """ return the cell cycle mean interpolated 
        from data existing after 16hAPF
        last experimental data point 27.97hAPF,
        after this np.interp takes the last known value
        of cell cycle time"""
        return np.interp(simulation_time - 16,
                         self.data_life_times_over_time[:,0],
                         self.data_life_times_over_time[:,1])
    def std_cell_cycle_function(self, simulation_time):
        """ return the cell cycle std,
        taken as a fraction of the mean after 16hAPF"""
        return self.COV_CELL_CYCLE * np.interp(simulation_time - 16,
                                               self.data_life_times_over_time[:, 0],
                                               self.data_life_times_over_time[:, 1])
    def cell_cycle_mean_std(self, simulation_time):
        """ return the cell cycle mean and std for any simulation time"""
        if simulation_time <= self.TRANSIENT_PHASE_STARTING_TIME:
            return [self.MEAN_CELL_CYCLE_BEGINNING,
                    self.STD_CELL_CYCLE_BEGINNING]
        if (simulation_time > self.TRANSIENT_PHASE_STARTING_TIME 
                and simulation_time <= self.SECOND_PHASE_STARTING_TIME):
            return [self.MEAN_CELL_CYCLE_TRANSIENT,
                    self.STD_CELL_CYCLE_TRANSIENT]
        if simulation_time > self.SECOND_PHASE_STARTING_TIME :
            if self.MODE == 1:
                return [self.mean_cell_cycle_function(simulation_time),
                        self.std_cell_cycle_function(simulation_time)]
            if self.MODE == 2:
                return [self.MU,
                        self.SIGMA]
        return np.nancl
    def cell_cycle_generator(self, simulation_time):
        """ return a cell cycle time taken from a log normal distribution,
        when the mean and std are known
        NOT USED HERE"""
        [mean_cell_cycle_time, std_cell_cycle_time] = self.cell_cycle_mean_std(simulation_time)
        mu = np.log(np.power(mean_cell_cycle_time, 2)/
                    np.sqrt(np.power(mean_cell_cycle_time, 2)+
                            np.power(std_cell_cycle_time, 2)))
        sigma = np.sqrt(np.log(1.0 + np.power(std_cell_cycle_time, 2)/
                               np.power(mean_cell_cycle_time, 2)))
        return log_normal_random(mu, sigma)
    def cell_cycle_generator_bivariate(self, simulation_time):
        """return a pair of life times from  a bivariate normal distribution"""
        [mean_cell_cycle_time, std_cell_cycle_time] = self.cell_cycle_mean_std(simulation_time)
        mean = [mean_cell_cycle_time, mean_cell_cycle_time]
        sigmasq = np.power(std_cell_cycle_time, 2)
        cov = [[sigmasq, self.RHO_CORRELATION_COEFFICIENT*sigmasq],
               [self.RHO_CORRELATION_COEFFICIENT*sigmasq, sigmasq]]
        x, y = np.random.multivariate_normal(mean, cov, 1).T
        return x, y

    def get_cell_types_after_division(self, simulation_time):
        """ functions which determines how to choose cell type after division
        at time simulation_tiime
        0 is a dividing cell, 1 an arrested cell, 2 a SOP"""
        #default return
        cell_types = [0, 0] #default return
        #random values which are used to make decisions
        random_value_1 = random.uniform(0, 1)
        random_value_2 = random.uniform(0, 1)
        random_value_3 = random.uniform(0, 1)
        random_value_4 = random.uniform(0, 1)

        if simulation_time > self.SECOND_PHASE_STARTING_TIME:
            if random_value_1 < hill_function(simulation_time,
                                              self.SWITCH_TIME_P,
                                              self.HILL_COEFF_P):
                if random_value_2 < hill_function(simulation_time,
                                                  self.SWITCH_TIME_ALPHA,
                                                  self.HILL_COEFF_ALPHA):
                    cell_types = [1, 1]
                else:
                    cell_types = [0, 1]
        #probability of giving rise to a SOP
        #only cells of type 0 can become SOPs
        #the probability PROBABILITY_SOP is conditioned on a cell
        #not being arrested
        if self.SWITCH_TIME_SOPs_ON < simulation_time < self.SWITCH_TIME_SOPs_OFF:
            #the first daughter cell can become a SOP
            if cell_types[0]==0 and random_value_3 < self.PROBABILITY_SOP:
                cell_types[0] = 2
            #the second daughter cell can become a SOP
            if cell_types[1]==0 and random_value_4 < self.PROBABILITY_SOP:
                cell_types[1] = 2
        return cell_types
    def run_population_growth_simulation(self):
        """actually run the simulation"""
        # ============================================================================
        # CREATE TISSUE
        # =============================================================================
        self.THIS_TISSUE = tissue(self.INITIAL_AGES,
                                  self.INITIAL_TYPES,
                                  self.INITIAL_CELL_CYCLE_TIMES,
                                  self.DIVIDING_CELL_TYPES,           
                                  self.cell_cycle_generator_bivariate,
                                  self.PAUSE_ON,
                                  self.PAUSE_OFF,
                                  self.AGE_ON,
                                  self.AGE_OFF,
                                  switch_type_after_division=1,
                                  get_cell_types_after_division=self.get_cell_types_after_division) 
        # =============================================================================
        # RUN POPULATION EVOLUTION
        # =============================================================================
        #start at 0 until SECOND_PHASE_STARTING_TIME
        self.THIS_TISSUE.run_population_evolution(self.STARTING_TIME,
                                                  self.SECOND_PHASE_STARTING_TIME,
                                                  self.DELTA_T)
        #at SECOND_PHASE_STARTING_TIME, split the tissue
        #into two parts which are simulated separately
        self.THIS_TISSUE_NO_BORDER = copy.deepcopy(self.THIS_TISSUE)
        self.THIS_TISSUE_NOT_NO_BORDER = copy.deepcopy(self.THIS_TISSUE)
        #at SECOND_PHASE_STARTING_TIME, split into two group of cells and run separately the two groups
        NO_BORDER_INITIAL_NUMBER_CELLS = 95 #manually set the initial number of noborder cells
        #first group
        self.THIS_TISSUE_NO_BORDER.cells = self.THIS_TISSUE_NO_BORDER.cells[:NO_BORDER_INITIAL_NUMBER_CELLS]
        #second group
        self.THIS_TISSUE_NOT_NO_BORDER.cells = self.THIS_TISSUE_NOT_NO_BORDER.cells[NO_BORDER_INITIAL_NUMBER_CELLS:]
        #run separately the two groups
        FIRST_TIME = self.SECOND_PHASE_STARTING_TIME + self.DELTA_T
        self.THIS_TISSUE_NO_BORDER.run_population_evolution(FIRST_TIME, self.TOTAL_TIME, self.DELTA_T)
        self.THIS_TISSUE_NOT_NO_BORDER.run_population_evolution(FIRST_TIME, self.TOTAL_TIME, self.DELTA_T)
        
        
    def plot_and_save_results(self):
        """ function to plot and save various results"""
        # =============================================================================
        # %% PLOT THE RESULT
        # =============================================================================
        #plot the number of cells
        plt.figure(1)
        plt.clf()
        #number of cells prior to 16hAPF
        nb_cells_0_array = np.array(self.THIS_TISSUE.nb_cells_array[0])
        nb_cells_1_array = np.array(self.THIS_TISSUE.nb_cells_array[1])
        nb_cells_2_array = np.array(self.THIS_TISSUE.nb_cells_array[2])
        nb_cells_all_array = nb_cells_0_array + nb_cells_1_array + nb_cells_2_array
        #number of cells in no border region, after 16hAPF
        nb_cells_0_no_border_array = np.array(self.THIS_TISSUE_NO_BORDER.nb_cells_array[0])
        nb_cells_1_no_border_array = np.array(self.THIS_TISSUE_NO_BORDER.nb_cells_array[1])
        nb_cells_2_no_border_array = np.array(self.THIS_TISSUE_NO_BORDER.nb_cells_array[2])
        nb_cells_no_border_all_array = (nb_cells_0_no_border_array
                                        + nb_cells_1_no_border_array
                                        + nb_cells_2_no_border_array)
        #total number of cells after 16hAPF
        nb_cells_0_total_array = (np.array(self.THIS_TISSUE_NOT_NO_BORDER.nb_cells_array[0])
                                  + nb_cells_0_no_border_array) 
        nb_cells_1_total_array = (np.array(self.THIS_TISSUE_NOT_NO_BORDER.nb_cells_array[1]) 
                                  + nb_cells_1_no_border_array)
        nb_cells_2_total_array = (np.array(self.THIS_TISSUE_NOT_NO_BORDER.nb_cells_array[2]) 
                                 + nb_cells_2_no_border_array)
        nb_cells_total_all_array = (nb_cells_0_total_array
                                    + nb_cells_1_total_array
                                    + nb_cells_2_total_array)
        
        plt.plot(self.THIS_TISSUE.time_array,
                 nb_cells_0_array,
                 label='dividing cells',
                 color='blue')
        plt.plot(self.THIS_TISSUE.time_array,
                 nb_cells_1_array,
                 label='arrested cells',
                 color='red')
        plt.plot(self.THIS_TISSUE.time_array,
                 nb_cells_2_array,
                 label='SOPs',
                 color='orange')
        plt.plot(self.THIS_TISSUE.time_array,
                 nb_cells_all_array,
                 label='all types',
                 color='green')
        plt.plot(self.THIS_TISSUE_NO_BORDER.time_array,
                 nb_cells_0_no_border_array,
                 color='blue', 
                 ls='--')
        plt.plot(self.THIS_TISSUE_NO_BORDER.time_array,
                 nb_cells_1_no_border_array,
                 color='red',
                 ls='--')
        plt.plot(self.THIS_TISSUE_NO_BORDER.time_array,
                 nb_cells_2_no_border_array,
                 color='orange', 
                 ls='--')
        plt.plot(self.THIS_TISSUE_NO_BORDER.time_array, 
                 nb_cells_no_border_all_array,
                 color='green', 
                 ls='--')
        plt.plot(self.THIS_TISSUE_NO_BORDER.time_array, 
                 nb_cells_0_total_array,
                 color='blue')
        plt.plot(self.THIS_TISSUE_NO_BORDER.time_array,
                 nb_cells_1_total_array,
                 color='red')
        plt.plot(self.THIS_TISSUE_NO_BORDER.time_array,
                 nb_cells_2_total_array,
                 color='orange')
        plt.plot(self.THIS_TISSUE_NO_BORDER.time_array,
                 nb_cells_total_all_array, 
                 color='green')
        plt.title('Total number of cells as a function of time \n dashed lines: no border region', fontsize=15)
        plt.xlabel('Time (hours)', fontsize=15)
        plt.ylabel('Number', fontsize=15)
        plt.axhline(y=0, color='black')
        plt.axvline(x=0, color='black')
        plt.box(False)
        plt.grid(True)
        plt.tick_params(axis='both', which='both', labelsize=15)
        plt.legend()
        plt.show()  
        
        ### save array with birth times and life times
        out = self.THIS_TISSUE_NO_BORDER.lifetimes_birthtimes_array
        np.savetxt(self.folder_name_results+"/LIFETIMES_ARRAY"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")
        
        ### save array with number of dividing cells, first phase
        out = np.column_stack((self.THIS_TISSUE.time_array,
                               nb_cells_0_array))
        np.savetxt(self.folder_name_results+"/nb_cells_first_phase_dividing_cells"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")
        ### save array with number of arrested cells, first phase
        out = np.column_stack((self.THIS_TISSUE.time_array,
                               nb_cells_1_array))
        np.savetxt(self.folder_name_results+"/nb_cells_first_phase_arrested_cells"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")        
        ### save array with number of SOPs, first phase
        out = np.column_stack((self.THIS_TISSUE.time_array, nb_cells_2_array))
        np.savetxt(self.folder_name_results+"/nb_cells_first_phase_SOPs"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")   
        ### save array with total number of cells, first phase
        out = np.column_stack((self.THIS_TISSUE.time_array,
                               nb_cells_all_array))
        np.savetxt(self.folder_name_results+"/nb_cells_first_phase_all_types"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")    
        ### save array with number of dividing cells in no border region, second phase
        out = np.column_stack((self.THIS_TISSUE_NO_BORDER.time_array,
                               nb_cells_0_no_border_array))
        np.savetxt(self.folder_name_results+"/nb_cells_second_phase_no_border_dividing_cells"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")
        ### save array with number of arrested cells in no border region, second phase
        out = np.column_stack((self.THIS_TISSUE_NO_BORDER.time_array,
                               nb_cells_1_no_border_array))
        np.savetxt(self.folder_name_results+"/nb_cells_second_phase_no_border_arrested_cells"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")        
       ### save array with number of SOP cells in no border region, second phase
        out = np.column_stack((self.THIS_TISSUE_NO_BORDER.time_array,
                               nb_cells_2_no_border_array))
        np.savetxt(self.folder_name_results+"/nb_cells_second_phase_no_border_SOPs"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")
       ### save array with total number of cells in no border region, second phase  
        out = np.column_stack((self.THIS_TISSUE_NO_BORDER.time_array,
                               nb_cells_no_border_all_array))
        np.savetxt(self.folder_name_results+"/nb_cells_second_phase_no_border_all_types"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ") 
         ### save array with number of dividing cells in full anterior region, second phase 
        out = np.column_stack((self.THIS_TISSUE_NO_BORDER.time_array,
                               nb_cells_0_total_array))
        np.savetxt(self.folder_name_results+"/nb_cells_second_phase_total_dividing_cells"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")
        ### save array with number of arrested cells in full anterior region, second phase 
        out = np.column_stack((self.THIS_TISSUE_NO_BORDER.time_array,
                               nb_cells_1_total_array))
        np.savetxt(self.folder_name_results+"/nb_cells_second_phase_total_arrested"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")  
        ### save array with number of SOP cells in full anterior region, second phase 
        out = np.column_stack((self.THIS_TISSUE_NO_BORDER.time_array,
                               nb_cells_2_total_array))
        np.savetxt(self.folder_name_results+"/nb_cells_second_phase_total_SOPs"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")       
        ### save array with total number of cells in full anterior region, second phase 
        out = np.column_stack((self.THIS_TISSUE_NO_BORDER.time_array,
                               nb_cells_total_all_array))
        np.savetxt(self.folder_name_results+"/nb_cells_second_phase_total_all_types"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")         
        
        ##calculate the rate of cell division, before and after smoothing
        nb_cells_all_joined_array = np.concatenate((nb_cells_all_array,
                                                    nb_cells_total_all_array))
        time_all_joined_array = np.concatenate((self.THIS_TISSUE.time_array,
                                                self.THIS_TISSUE_NO_BORDER.time_array))
        kd_array = np.diff(nb_cells_all_joined_array)/(map(float,nb_cells_all_joined_array[:-1]))/self.DELTA_T
        kd_array_smoothed = moving_average(time_all_joined_array,
                                           kd_array )
        kd_array_no_border = np.diff(nb_cells_no_border_all_array)/map(float,nb_cells_no_border_all_array[:-1])/self.DELTA_T
        kd_array_no_border_smoothed = moving_average(self.THIS_TISSUE_NO_BORDER.time_array[:-1],
                                                     kd_array_no_border )
        plt.figure(2)
        plt.clf()
        plt.plot(kd_array_smoothed[:,0],
                 kd_array_smoothed[:,1],
                 label='all cells',
                 lw=3)
        plt.plot(kd_array_no_border_smoothed[:,0],
                 kd_array_no_border_smoothed[:,1],
                 label='no border cells',
                 lw=3)
        plt.title('Rate of cell division as a function of time',
                  fontsize=15)
        plt.xlabel('Time (hours)',
                   fontsize=15)
        plt.ylabel('Rate (hours -1)',
                   fontsize=15)
        plt.axhline(y=0, color='black')
        plt.axvline(x=0, color='black')
        plt.box(False)
        plt.grid(True)
        plt.tick_params(axis='both',
                        which='both',
                        labelsize=15)
        plt.legend()
        plt.show()
        
        
        ## save result for cell division rate in CSV files
        divs_frame_array = self.THIS_TISSUE_NO_BORDER.divs_frame_array
        np.savetxt(self.folder_name_results+"/divs_frame_array"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   divs_frame_array,
                   delimiter = " ")
        out = kd_array_smoothed
        np.savetxt(self.folder_name_results+"/kd_array_smoothed"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")
        out = kd_array_no_border_smoothed
        np.savetxt(self.folder_name_results+"/kd_array_no_border_smoothed"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")       
        out = np.column_stack([time_all_joined_array[:-1], kd_array])
        np.savetxt(self.folder_name_results+"/kd_array"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")
        out = np.column_stack([self.THIS_TISSUE_NO_BORDER.time_array[:-1], kd_array_no_border])
        np.savetxt(self.folder_name_results+"/kd_array_no_border"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")
        
        
        # =============================================================================
        # %% NUMBER OF CELLS PRIOR TO 16hAPF
        # =============================================================================
        plt.figure(3)
        plt.clf()
        plt.plot(self.THIS_TISSUE.time_array,
                 nb_cells_all_array,
                 label='simulation',
                 color='green',
                 lw=3)
        plt.plot(self.THIS_TISSUE_NO_BORDER.time_array,
                 nb_cells_total_all_array,
                 color='green',
                 lw=3)
        plt.title('Number of cells as a function of time, prior to 16hAPF', fontsize=15)
        plt.xlabel('Time (hours)', fontsize=15)
        plt.ylabel('Number', fontsize=15)
        plt.axhline(y=0, color='black')
        plt.axvline(x=0, color='black')
        plt.box(False)
        plt.grid(True)
        plt.tick_params(axis='both', which='both', labelsize=15)
        plt.legend()
        plt.show()
                
        
        # =============================================================================
        # %% NUMBER OF CELLS AFTER 16hAPF
        # =============================================================================
                
        fig = plt.figure(4)
        plt.clf()
        fig.set_size_inches(8,4)
        [ax1, ax2] = fig.subplots(1,2,sharex=True, sharey=True)
        plt.subplots_adjust(wspace=1.0, hspace=0.8)
        #plot simulation  resultsfor fulla nterior
        ax1.plot(self.THIS_TISSUE.time_array,
                 nb_cells_0_array,
                 label='dividing cells',
                 color='blue',
                 ls='--')
        ax1.plot(self.THIS_TISSUE.time_array,
                 nb_cells_1_array,
                 label='arrested cells',
                 color='red',
                 ls='--')
        ax1.plot(self.THIS_TISSUE.time_array,
                 nb_cells_2_array,
                 label='SOPs',
                 color='orange',
                 ls='--')
        ax1.plot(self.THIS_TISSUE.time_array,
                 nb_cells_all_array,
                 label='all types',
                 color='green',
                 ls='--')
        ax1.plot(self.THIS_TISSUE_NO_BORDER.time_array,
                 nb_cells_0_total_array,
                 color='blue',
                 ls='--')
        ax1.plot(self.THIS_TISSUE_NO_BORDER.time_array,
                 nb_cells_1_total_array,
                 color='red',
                 ls='--')
        ax1.plot(self.THIS_TISSUE_NO_BORDER.time_array,
                 nb_cells_2_total_array,
                 color='orange',
                 ls='--')
        ax1.plot(self.THIS_TISSUE_NO_BORDER.time_array,
                 nb_cells_total_all_array,
                 color='green',
                 ls='--')
        #simulation: no border
        ax2.plot(self.THIS_TISSUE_NO_BORDER.time_array,
                 nb_cells_0_no_border_array,
                 color='blue',
                 ls='--')
        ax2.plot(self.THIS_TISSUE_NO_BORDER.time_array,
                 nb_cells_1_no_border_array,
                 color='red',
                 ls='--')
        ax2.plot(self.THIS_TISSUE_NO_BORDER.time_array,
                 nb_cells_2_no_border_array,
                 color='orange',
                 ls='--')
        ax2.plot(self.THIS_TISSUE_NO_BORDER.time_array,
                 nb_cells_no_border_all_array,
                 color='green',
                 ls='--')
        ax1.set_title('total number of cells anterior,\n simulation', fontsize=15)
        ax2.set_title('total number of cells no border,\n simulation', fontsize=15)
        ax1.set_xlabel('Time (hours)', fontsize=15)
        ax2.set_xlabel('Time (hours)', fontsize=15)
        ax1.set_ylabel('Number', fontsize=15)
        ax2.set_ylabel('Number', fontsize=15)
        ax1.grid('true')
        ax2.grid('true')
        ax1.tick_params(axis='both', which='both', labelsize=15, labelleft=True)
        ax2.tick_params(axis='both', which='both', labelsize=15, labelleft=True)
        plt.tight_layout()

        
        # =============================================================================
        # %% PLOT CELL CYCLE TIMES THAT WERE USED AS INPUT
        # =============================================================================
        plt.figure(5)
        plt.clf()
        #cell cycles first phase
        cell_cycles_mean_0 =[self.cell_cycle_mean_std(this_time)[0] 
                             for this_time
                             in self.THIS_TISSUE.time_array]
        cell_cycles_std_0 = [self.cell_cycle_mean_std(this_time)[1]
                              for this_time
                              in self.THIS_TISSUE.time_array]
        #cell cycles second phase in no border region
        cell_cycles_mean_1 = [self.cell_cycle_mean_std(this_time)[0]
                              for this_time
                              in self.THIS_TISSUE_NO_BORDER.time_array]
        cell_cycles_std_1 = [self.cell_cycle_mean_std(this_time)[1]
                             for this_time
                             in self.THIS_TISSUE_NO_BORDER.time_array]
        
        ## save CSV files for cell cycle times
        out = np.column_stack((self.THIS_TISSUE.time_array,
                               cell_cycles_mean_0,
                               cell_cycles_std_0))
        np.savetxt(self.folder_name_results+"/cell_cycles_mean_std_first_phase"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")
        
        out = np.column_stack((self.THIS_TISSUE_NO_BORDER.time_array,
                               cell_cycles_mean_1,
                               cell_cycles_std_1))
        np.savetxt(self.folder_name_results+"/cell_cycles_mean_std_second_phase"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")

        # plot the cell cycle times that were chosen
        plt.plot(self.THIS_TISSUE.time_array, cell_cycles_mean_0 , color='blue', lw=3)
        plt.plot(self.THIS_TISSUE.time_array, np.array(cell_cycles_mean_0) + np.array(cell_cycles_std_0), color='blue', lw=1)
        plt.plot(self.THIS_TISSUE.time_array, np.array(cell_cycles_mean_0) - np.array(cell_cycles_std_0), color='blue', lw=1)
        plt.plot(self.THIS_TISSUE_NO_BORDER.time_array, cell_cycles_mean_1, color='blue', lw=3)
        plt.plot(self.THIS_TISSUE_NO_BORDER.time_array, np.array(cell_cycles_mean_1) + np.array(cell_cycles_std_1), color='blue',lw=1)
        plt.plot(self.THIS_TISSUE_NO_BORDER.time_array, np.array(cell_cycles_mean_1) - np.array(cell_cycles_std_1), color='blue',lw=1)
        
        plt.errorbar(self.data_life_times_over_time[:,0] + 16,
                     self.data_life_times_over_time[:,1],
                     self.data_life_times_over_time[:,2],
                     fmt='o',
                     ls='none',
                     lw=3,
                     capsize=3,
                     color='blue')
        plt.ylim([0, 9.0])
        plt.title('cell cycle time, input from experimental measurements')
        plt.xlabel('Time (hours)')
        plt.ylabel('Time (hours)')
        plt.axhline(y=0, color='black')
        plt.axvline(x=0, color='black')
        plt.box(False)
        plt.grid(True)
        plt.legend()
        
        

# =============================================================================
# ADDITIONAL MATHEMATICAL FUNCTIONS
# =============================================================================
def log_normal_random(mu, sigma):
    """return a variable with log-normal distribution"""
    return np.exp(mu+sigma*random.gauss(0, 1))
def hill_function(time, switch_time, hill_coeff,):
    """convenient function for the p probability distribution"""
    val = np.power(time/switch_time,hill_coeff)
    return val / (1.0 + val) 
 #define a function that performs smoothing
N_moving = 21 #size of moving average window 
def moving_average(time_array, array):
    array_convolved_time = np.convolve(time_array,
                                       np.ones((N_moving,))/N_moving,
                                       mode='valid')
    array_convolved = np.convolve(array,
                                  np.ones((N_moving,))/N_moving,
                                  mode='valid')
    return np.array([[array_convolved_time[i], array_convolved[i]]
                     for i
                     in range(array_convolved.shape[0])])