#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Wed Jan  2 17:21:14 2019

@author: salbreg/williamsonj
"""

#!/usr/bin/python
import numpy as np
from Cell import cell

#----------------------------------------------------------
# Class Tissue
#----------------------------------------------------------
class tissue(object):
    """ This class contains objects and methods to manipulate the tissue""" ## JJW CHANGES
    def __init__(self, 
                 ages,
                 types,
                 cell_cycle_times,
                 dividing_cell_types, 
                 cell_cycle_generator,
                 PAUSE_ON,
                 PAUSE_OFF,
                 AGE_ON,
                 AGE_OFF,
                 **kwargs):
        """ initialize the simulations
        ages is the list of ages of cells at the beginning of the simulation
        types is the list of types of cells at the beginning of the simulation
        types is the list of cell cycle time of cells at the beginning of the simulation
        dividing_cell_types is a dictionary which indicates if cell types are dividing 
        (indicated by a 1 in the dictionary)
        or not (indicated by a 0 in the dictionary)
        cell_cycle_generator returns two lifetimes for two daughter cells 
        as a function of the simulation_time
        PAUSE_ON, PAUSE_OFF, AGE_ON, AGE_OFF are parameters
        for phases where either cells do not divide (PAUSE)
        or do not age (AGE)
        """
        #list of cells
        self.cells = [cell(ages[k], types[k], cell_cycle_times[k]) for k, _ in enumerate(ages)]
        #last cell index - used to create new keys for dictionaries
        self.last_cell_index = len(self.cells)-1
        self.switch_type_after_division = kwargs.get('switch_type_after_division', 0)
        self.get_cell_types_after_division = kwargs.get('get_cell_types_after_division', 0)
        self.cell_cycle_generator = cell_cycle_generator
        self.dividing_cell_types = dividing_cell_types
        self.number_types = len(dividing_cell_types) #the length of this dictionary is the number of types    
        self.PAUSE_ON = PAUSE_ON
        self.PAUSE_OFF = PAUSE_OFF
        self.AGE_ON = AGE_ON
        self.AGE_OFF = AGE_OFF
        # used to store the result
        self.nb_cells_array = {}
        self.divs_frame_array = []  ## counter for the number of divisions in each frame. 
        self.lifetimes_birthtimes_array = [] ## store lifetimes and birth times and cell type at each division
        
    def run_population_evolution(self, FIRST_TIME, TOTAL_TIME, DELTA_T):
        """ run and record the evolution of the population
        the number of cells of each type as a function of time is saved in a dictionary, nb_cells_array
        """
        self.time_array = np.linspace(FIRST_TIME,
                                      TOTAL_TIME,
                                      1+(TOTAL_TIME-FIRST_TIME)/DELTA_T)
        dt = self.time_array[1] - self.time_array[0]
        self.nb_cells_array = {}
        
        ## counter for the number of divisions in each frame. 
        self.divs_frame_array = []

        
        self.lifetimes_birthtimes_array = []
        for k in range(self.number_types):
            self.nb_cells_array[k] = []
        #run simulation
        for this_time in self.time_array:
            #at each time save the number of cells of different types
            for k in range(self.number_types):
                cells_of_type_k = [this_cell
                                   for this_cell
                                   in self.cells
                                   if this_cell.type==k]
                self.nb_cells_array[k].append(len(cells_of_type_k)) 
            #run one time point of simulation
            self.advance_cell_age_and_divide(this_time, dt)
    def advance_cell_age_and_divide(self,
                                    this_time,
                                    dt,
                                    **kwargs):
        """advance every cell age by dt,
        and divide cells that have reached their cell cycle time
        """
        
        #counter for the number of divisions that are performed
        #within this function
        howmanydivs = 0 
        
        for this_cell_index in range(len(self.cells)):
            this_cell =  self.cells[this_cell_index]
            
            if this_time > self.AGE_ON or this_time <= self.AGE_OFF:
                this_cell.age += dt
            if (this_cell.age > this_cell.cell_cycle_time+(0.5*dt)
                and self.dividing_cell_types[this_cell.type] 
                and (this_time < self.PAUSE_ON or this_time > self.PAUSE_OFF)):
                #DIVISION EVENT
                howmanydivs = howmanydivs + 1
                this_cell.age = 0
                daughter_cell_type = this_cell.type
                if self.switch_type_after_division:
                    types = self.get_cell_types_after_division(this_time)
                    this_cell.type = types[0]
                    daughter_cell_type = types[1]
                daughter_cell_age = 0
                # call cell cycle generator to generate two lifetimes for the sisters
                #these two lifetimes can be correlated
                life_time_1, life_time_2 = self.cell_cycle_generator(this_time)  
                    
                daughter_cell = cell(daughter_cell_age,
                                     daughter_cell_type,
                                     life_time_1) 
                #the mother cell becomes ones of the daughter cells
                this_cell.cell_cycle_time = life_time_2
                self.cells.append(daughter_cell)
                self.last_cell_index += 1
                    
                ## save information about the division event
                self.lifetimes_birthtimes_array.append([this_time,
                                                        life_time_1,
                                                        daughter_cell_type,
                                                        life_time_2,
                                                        this_cell.type])
                
        ## all divisions have taken place so append a row to divs_frame_array.
        self.divs_frame_array.append([this_time, howmanydivs]) 