#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 25 13:39:56 2019

@author: salbreg/williamsonj
"""

import random
import numpy as np
import matplotlib.pyplot as plt
from Tissue import tissue


class run_population_growth(object):
    """ class that run a simulation for population growth"""
    DIVIDING_CELL_TYPES = {0:1, 1:0, 2:0}#cells of type 0 divide, cells of type 1 or 2 (SOPs) do not
    DELTA_T = 2.5/60.0 #as in experiment, delta_t is taken to be 2.5 minutes
    TOTAL_TIME = 30 #total time in hours for the simulation
    def __init__(self,
                 file_suffix,
                 STARTING_TIME,
                 INITIAL_NUMBER_CELLS,
                 INITIAL_AGES,
                 MU,
                 SIGMA,
                 RHO_CORRELATION_COEFFICIENT,
                 MODE,
                 folder_name_results):
        """class initialization
        MODE can be 1 (mean of cell cycle time distribution in second phase
        is taken from a fit to experimental data  and std is proportional to mean)
        or can be 2 ( cell cycle time distribution has a fixed mean and std)"""
        self.file_suffix = file_suffix
        self.STARTING_TIME = STARTING_TIME
        self.INITIAL_NUMBER_CELLS = INITIAL_NUMBER_CELLS
        self.MU = MU
        self.SIGMA = SIGMA
        self.RHO_CORRELATION_COEFFICIENT = RHO_CORRELATION_COEFFICIENT
        self.MODE = MODE
        self.folder_name_results = folder_name_results
        #initial cells have an age INITIAL_AGES
        self.INITIAL_AGES = INITIAL_AGES
        #initial cells have cell type 0 by default
        self.INITIAL_TYPES = np.zeros(self.INITIAL_NUMBER_CELLS) 
        #generate a set of initial cell cycle times
        self.INITIAL_CELL_CYCLE_TIMES = np.zeros(len(INITIAL_AGES))
#note: beyond the boundaries of discrete values,
# the interp function returns the closest value
    def mean_cell_cycle_function(self, simulation_time):
        """ return the cell cycle mean interpolated 
        from data existing after 16hAPF
        last experimental data point 27.97hAPF,
        after this np.interp takes the last known value
        of cell cycle time"""
        return np.interp(simulation_time - 16,
                         self.data_life_times_over_time[:,0],
                         self.data_life_times_over_time[:,1])
    def std_cell_cycle_function(self, simulation_time):
        """ return the cell cycle std,
        taken from experimental measurements"""
        return np.interp(simulation_time - 16,
                         self.data_life_times_over_time[:,0],
                         self.data_life_times_over_time[:,2])
    def cell_cycle_mean_std(self, simulation_time):
        """ return the cell cycle mean and std for any simulation time"""

        if self.MODE == 1:
            return [self.mean_cell_cycle_function(simulation_time),
                    self.std_cell_cycle_function(simulation_time)]
        if self.MODE == 2:
            return [self.MU, self.SIGMA]
        return np.nan
    def cell_cycle_generator(self, simulation_time):
        """ return a cell cycle time taken from a log normal distribution,
        when the mean and std are known
        NOT USED HERE"""
        [mean_cell_cycle_time, std_cell_cycle_time] = self.cell_cycle_mean_std(simulation_time)
        mu = np.log(np.power(mean_cell_cycle_time, 2)/
                    np.sqrt(np.power(mean_cell_cycle_time, 2)+
                            np.power(std_cell_cycle_time, 2)))
        sigma = np.sqrt(np.log(1.0 + np.power(std_cell_cycle_time, 2)/
                               np.power(mean_cell_cycle_time, 2)))
        return log_normal_random(mu, sigma)
    def cell_cycle_generator_bivariate(self, simulation_time):
        """return a pair of life times from  a bivariate normal distribution"""
        [mean_cell_cycle_time,
         std_cell_cycle_time] = self.cell_cycle_mean_std(simulation_time)
        mean = [mean_cell_cycle_time, mean_cell_cycle_time]
        sigmasq = np.power(std_cell_cycle_time, 2)
        cov = [[sigmasq,
                self.RHO_CORRELATION_COEFFICIENT*sigmasq],
               [self.RHO_CORRELATION_COEFFICIENT*sigmasq,
                sigmasq]]
        x, y = np.random.multivariate_normal(mean, cov, 1).T
        return x, y
    def get_cell_types_after_division(self, simulation_time):
        """ functions which determines how to choose cell type after division
        at time simulation_tiime
        0 is a dividing cell, 1 an arrested cell, 2 a SOP"""
        #default return
        cell_types = [0, 0] #default return   
        return cell_types
    def run_population_growth_simulation(self):
        """actually run the simulation"""
        # ============================================================================
        # CREATE TISSUE
        # =============================================================================
        self.THIS_TISSUE = tissue(self.INITIAL_AGES,
                                  self.INITIAL_TYPES,
                                  self.INITIAL_CELL_CYCLE_TIMES,
                                  self.DIVIDING_CELL_TYPES,    
                                  self.cell_cycle_generator_bivariate,
                                  switch_type_after_division=1,
                                  get_cell_types_after_division=self.get_cell_types_after_division) 
        # =============================================================================
        # RUN POPULATION EVOLUTION
        # =============================================================================
        #start at STARTING_TIME until TOTAL_TIME
        self.THIS_TISSUE.run_population_evolution(self.STARTING_TIME,
                                                  self.TOTAL_TIME,
                                                  self.DELTA_T)                       
    def plot_and_save_results(self):
        """ function to plot and save various results"""
        # =============================================================================
        # %% PLOT THE RESULT
        # =============================================================================
        #plot the number of cells
        plt.figure(1)
        plt.clf()
        #number of cells
        nb_cells_0_array = np.array(self.THIS_TISSUE.nb_cells_array[0])
        nb_cells_1_array = np.array(self.THIS_TISSUE.nb_cells_array[1])
        nb_cells_2_array = np.array(self.THIS_TISSUE.nb_cells_array[2])
        nb_cells_all_array = nb_cells_0_array + nb_cells_1_array + nb_cells_2_array
        
        plt.plot(self.THIS_TISSUE.time_array, nb_cells_0_array, label='dividing cells', color='blue')
        plt.plot(self.THIS_TISSUE.time_array, nb_cells_1_array, label='arrested cells' , color='red')
        plt.plot(self.THIS_TISSUE.time_array, nb_cells_2_array, label='SOPs' , color='orange')
        plt.plot(self.THIS_TISSUE.time_array, nb_cells_all_array, label='all types', color='green')
        plt.title('Total number of cells as a function of time \n dashed lines: no border region', fontsize=15)
        plt.xlabel('Time (hours)', fontsize=15)
        plt.ylabel('Number', fontsize=15)
        plt.axhline(y=0, color='black')
        plt.axvline(x=0, color='black')
        plt.box(False)
        plt.grid(True)
        plt.tick_params(axis='both', which='both', labelsize=15)
        plt.legend()
        
        ### save array with birth times and life times
        out = self.THIS_TISSUE.lifetimes_birthtimes_array
        np.savetxt(self.folder_name_results+"/LIFETIMES_ARRAY"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")
        
        ### save array with number of dividing cells
        out = np.column_stack((self.THIS_TISSUE.time_array,
                               nb_cells_0_array))
        np.savetxt(self.folder_name_results+"/nb_cells_dividing_cells"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")
        ### save array with number of arrested cells
        out = np.column_stack((self.THIS_TISSUE.time_array,
                               nb_cells_1_array))
        np.savetxt(self.folder_name_results+"/nb_cells_arrested_cells"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")        
        ### save array with number of SOPs
        out = np.column_stack((self.THIS_TISSUE.time_array, nb_cells_2_array))
        np.savetxt(self.folder_name_results+"/nb_cells_SOPs"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")   
        ### save array with total number of cells
        out = np.column_stack((self.THIS_TISSUE.time_array,
                               nb_cells_all_array))
        np.savetxt(self.folder_name_results+"/nb_cells_all_types"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")    
     
        ##calculate the rate of cell division, before and after smoothing
        kd_array = np.diff(nb_cells_all_array)/(map(float,nb_cells_all_array[:-1]))/self.DELTA_T
        kd_array_smoothed = moving_average(self.THIS_TISSUE.time_array[:-1],
                                           kd_array )
        plt.figure(2)
        plt.clf()
        plt.plot(kd_array_smoothed[:,0],
                 kd_array_smoothed[:,1],
                 label='all cells',
                 lw=3)
        plt.title('Rate of cell division as a function of time',
                  fontsize=15)
        plt.xlabel('Time (hours)',
                   fontsize=15)
        plt.ylabel('Rate (hours -1)',
                   fontsize=15)
        plt.axhline(y=0, color='black')
        plt.axvline(x=0, color='black')
        plt.box(False)
        plt.grid(True)
        plt.tick_params(axis='both',
                        which='both',
                        labelsize=15)
        plt.legend()
        
        
        ## save result for cell division rate in CSV files
        divs_frame_array = self.THIS_TISSUE.divs_frame_array
        np.savetxt(self.folder_name_results+"/divs_frame_array"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   divs_frame_array,
                   delimiter = " ")
        out = kd_array_smoothed
        np.savetxt(self.folder_name_results+"/kd_array_smoothed"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")
        out = np.column_stack([self.THIS_TISSUE.time_array[:-1], kd_array])
        np.savetxt(self.folder_name_results+"/kd_array"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")

        
        
        # =============================================================================
        # %% PLOT NUMBER OF CELLS 
        # =============================================================================
                
        plt.figure(4)
        plt.clf()
        #plot simulation  resultsfor fulla nterior
        plt.plot(self.THIS_TISSUE.time_array,
                 nb_cells_0_array,
                 label='dividing cells',
                 color='blue',
                 ls='--')
        plt.plot(self.THIS_TISSUE.time_array,
                 nb_cells_1_array,
                 label='arrested cells',
                 color='red',
                 ls='--')
        plt.plot(self.THIS_TISSUE.time_array,
                 nb_cells_2_array,
                 label='SOPs',
                 color='orange',
                 ls='--')
        plt.plot(self.THIS_TISSUE.time_array,
                 nb_cells_all_array,
                 label='all types',
                 color='green',
                 ls='--')
        plt.title('total number of cells anterior,\n simulation', fontsize=15)
        plt.xlabel('Time (hours)', fontsize=15)
        plt.ylabel('Number', fontsize=15)
        plt.grid('true')
        plt.tick_params(axis='both', which='both', labelsize=15, labelleft=True)
        plt.tight_layout()

        
        # =============================================================================
        # %% PLOT CELL CYCLE TIMES THAT WERE USED AS INPUT
        # =============================================================================
        plt.figure(5)
        plt.clf()
        #cell cycles first phase
        cell_cycles_mean_0 =[self.cell_cycle_mean_std(this_time)[0] 
                             for this_time
                             in self.THIS_TISSUE.time_array]
        cell_cycles_std_0 = [self.cell_cycle_mean_std(this_time)[1]
                              for this_time
                              in self.THIS_TISSUE.time_array]
        
        ## save CSV files for cell cycle times
        out = np.column_stack((self.THIS_TISSUE.time_array,
                               cell_cycles_mean_0,
                               cell_cycles_std_0))
        np.savetxt(self.folder_name_results+"/cell_cycles_mean_std_first_phase"
                   +"_run_"
                   +self.file_suffix
                   +".csv",
                   out,
                   delimiter = " ")
        
       

        # plot the cell cycle times that were chosen
        plt.plot(self.THIS_TISSUE.time_array, cell_cycles_mean_0 , color='blue', lw=3)
        plt.plot(self.THIS_TISSUE.time_array, np.array(cell_cycles_mean_0) + np.array(cell_cycles_std_0), color='blue', lw=1)
        plt.plot(self.THIS_TISSUE.time_array, np.array(cell_cycles_mean_0) - np.array(cell_cycles_std_0), color='blue', lw=1)
          
       
        

# =============================================================================
# ADDITIONAL MATHEMATICAL FUNCTIONS
# =============================================================================
def log_normal_random(mu, sigma):
    """return a variable with log-normal distribution"""
    return np.exp(mu+sigma*random.gauss(0, 1))
def hill_function(time, switch_time, hill_coeff,):
    """convenient function for the p probability distribution"""
    val = np.power(time/switch_time,hill_coeff)
    return val / (1.0 + val) 
 #define a function that performs smoothing
N_moving = 21 #size of moving average window 
def moving_average(time_array, array):
    array_convolved_time = np.convolve(time_array,
                                       np.ones((N_moving,))/N_moving,
                                       mode='valid')
    array_convolved = np.convolve(array,
                                  np.ones((N_moving,))/N_moving,
                                  mode='valid')
    return np.array([[array_convolved_time[i], array_convolved[i]]
                     for i
                     in range(array_convolved.shape[0])])